/******************************************************************************
* To the extent possible under law, Sebastian Roith
* <https://www.roith-unterrichtet.de> has waived all copyright and related or
* neighboring rights to this script. This work is published from Germany.
******************************************************************************/

/**
* Klasse, die den Spielablauf eines Spiels, beim anhand einer Klassenkarte eine Objektkarte auszufüllen ist, verwaltet, die Kommunikation mit der Lernplattform über SCORM-API übernimmt und die Anzeige der nötigen Elemente sicherstellt
*/
function ReadClasscardTask(
    nodeParent,     // DOM-Knoten, als dessen einzige Kinder das Spiel angezeigt werden soll
    classes,        // Verfügbare Klassen
    targetConfig    // Erwünschte Konfiguration
) {
    ReadClasscardTask.parent.call(this);
    this.msgPlace = new ScoTools.MessagePlace();
    this.msgPlace.insertThisBefore(nodeParent.getElementsByTagName('div')[0]);
    this.msgPlaceGeneral = new ScoTools.MessagePlace(true);
    this.msgPlaceGeneral.insertThisBefore(nodeParent.firstChild);
    this.nodeParent = nodeParent;
    this._classes = classes;

    this.connect();
    this.establish();
    this.startLevel(targetConfig);
    AudioFeedbackTask.cache();
    this.ready = true;
}
ReadClasscardTask.prototype = Object.create(AudioFeedbackTask.prototype);
ReadClasscardTask.prototype.constructor = ReadClasscardTask;
ReadClasscardTask.parent = AudioFeedbackTask;

// Ist das Spiel gestartet?
ReadClasscardTask.prototype.ready = false;

// Zu erreichende Konfiguration
ReadClasscardTask.prototype.targetConfig = null;

// Bereits erfolgreich absolvierte Level
ReadClasscardTask.prototype.countDone = 0;

// Zu absolvierende Levelanzahl
ReadClasscardTask.prototype.countNeeded = 1;

/**
* Ein Level starten
* @return undefined
*/
ReadClasscardTask.prototype.startLevel = function(
    targetConfig    // Erwünschte Konfiguration
) {
    this.targetConfig = (targetConfig instanceof Array) ? targetConfig : [ targetConfig ];
    this._nodeGoal = this._change(this.targetConfig[0].className, this.targetConfig[0], this._nodeGoal, this._nodeGoalParent);
    this.procChange();
}

/**
* Erstellt einen DOM-Knoten passend zum Klassennamen classNameanhand von Attributwerten und ersetzt nodeOld durch den neu erstellten Knoten bzw. fügt ihn am Ende von parentNode ein
* @return SVGElement    Erstellter DOM-Knoten
*/
ReadClasscardTask.prototype._change = function(
    className,  // Klassenname der gewünschten Klasse
    atts,       // Objekt mit den tatsächlichen Attributwerten
    nodeOld,    // Zu ersetzender DOM-Knoten
    parentNode  // Übergeordneter DOM-Knoten
) {
    var classIndex =  this._classes.map(function(item) { return item.className; }).indexOf(className);
    if (classIndex == -1) {
        if (nodeOld && nodeOld.parentNode)
            nodeOld.parentNode.removeChild(nodeOld);
        return null;
    }
    return this._classes[classIndex].viewer.change(atts, nodeOld, parentNode);
}

/**
* Aus Titelzeile der Objektkarte Objekt- und Klassenname extrahieren
* @return Object    { klasse: String, objekt: String, error: "String oder undefined" }
*/
ReadClasscardTask.parseTitle = function(
    s       // Textinhalt der Titelzeile
) {
    var parts = s.split(':');
    switch (parts.length) {
        case 1:
            return { klasse: '', objekt: s.trim(), error: 'In Zeile 1 fehlt der Doppelpunkt.' };
        case 2:
            break;
        default:
            return { klasse: '', objekt: s.trim(), error: 'In Zeile 1 kann es nur einen Doppelpunkt geben.' };
    }
    var klasse = parts[1].trim();
    var objekt = parts[0].trim();
    if (!klasse.length)
        return { klasse: klasse, objekt: objekt, error: 'In Zeile 1 fehlt der Klassenname.' };
    var regVerboten = new RegExp('\\W');
    var arrVerboten = regVerboten.exec(objekt);
    if (arrVerboten && arrVerboten.length)
        return { klasse: klasse, objekt: objekt, error: 'Objektname in Zeile 1 enthält verbotenes Zeichen: ' + arrVerboten[0] };
    arrVerboten = regVerboten.exec(klasse);
    if (arrVerboten && arrVerboten.length)
        return { klasse: klasse, objekt: objekt, error: 'Klassenname in Zeile 1 enthält verbotenes Zeichen: ' + arrVerboten[0] };
    return { klasse: klasse, objekt: objekt };
}

/**
* Aus Objektkarte Attribute auslesen
* @return mixed    Gelesene Werte als Objekt oder String mit Fehlermeldung
*/
ReadClasscardTask.parseAttributes = function(
    s,          // Attributaufzählung auf Objektkarte
    arrValid    // Liste mit erlaubten Attributnamen
) {
    var obj = {};
    var arrInfos = s.split('\n');
    var s, parts, attname, attvalue;
    for (var z = 0; z < arrInfos.length; ++z) {
        s = arrInfos[z].trim();
        if (!s.length)
            continue;
        parts = s.split('=');
        switch (parts.length) {
            case 1:
                return 'Gleichheitszeichen fehlt in Zeile ' + (z + 2) + '.';
            case 2:
                break;
            default:
                return 'In Zeile ' + (z + 2) + ' kann es nicht mehrere Gleichheitszeichen geben.';
        }
        attname = parts[0].trim();
        attvalue = parts[1].trim();
        if (arrValid.indexOf(attname) == -1)
            return 'Fehler in Zeile ' + (z + 2) + ': Ein Attribut namens „' + attname + '“ ist nicht vorgesehen.';
        if (!attvalue.length)
            return 'In Zeile ' + (z + 2) + ' fehlt beim Attribut „' + attname + '“ der Attributwert.';
        if (obj[attname] !== undefined)
            return 'Zum Attribut „' + attname + '“ darf nicht mehrfach ein Wert angegeben werden.';
        obj[attname] = attvalue;
    }
    return obj;
}

/**
* Ermittelt die gespeicherte Eingabe
* @return Object    Objekt der Form { title: "Inhalt der obersten Zeile der Objektkarte", atts: "Inhalt des unteren Bereichs der Objektkarte" }
*/
ReadClasscardTask.prototype._getSaved = function() {
    if (this.isConnected()) {
        var data = this.read('suspendData');
        if (data) {
            data = data.split('\n');
            return { title: data.length ? data.shift() : '', atts: data.join('\n') };
        }
    }
    this.countDone = this.getDone() ? 1 : 0;
    return { title: '', atts: '' };
}

/**
* Aktuellen Status speichern
* @return undefined
*/
ReadClasscardTask.prototype.doSave = function() {
    if (this.getDone()) {
        this.reportTime();
        return;
    }
    if (!this.lmslink.ready)
    if (!this.lmslink.ready)
        return;
    this.save({
        'score': (this.countDone >= this.countNeeded) ? 100 : 0,
        'scoreMin': 0,
        'scoreMax': 100,
        'lessonStatus': (this.countDone >= this.countNeeded) ? 'completed' : 'incomplete',
        'suspendData': this._okt.value + '\n' + this._oka.value
    });
}

/**
* Zur Objektkarte passendes Objekt anzeigen
* @return undefined
*/
ReadClasscardTask.prototype.showObject = function(
    className,      // Interner Klassenname der Klasse
    objAtts         // Objekt mit tatsächlichen Attributen
) {
    this._nodeNow = this._change(className, objAtts, this._nodeNow, this._nodeNowParent);
}

/**
* Bedienelemente für das Lernpaket aufbauen
* @return undefined
*/
ReadClasscardTask.prototype.establish = function() {
    ReadClasscardTask.parent.prototype.establish.call(this);
    this._nodeGoal = this.nodeParent.ownerDocument.getElementById('zielobjekt');
    this._nodeGoalParent = this._nodeGoal.parentNode;
    this._nodeNow = this.nodeParent.ownerDocument.getElementById('aktuell');
    this._nodeNowParent = this._nodeNow.parentNode;
    var forms = this.nodeParent.getElementsByTagName('form');
    var self = this;
    var data = this._getSaved();
    var classes;
    for (var n = 0; n < forms.length; ++n) {
        forms[n].addEventListener('submit', function(ev) { ev.preventDefault; ev.stopPropagation });
        classes = forms[n].className.split(' ');
        if (classes.indexOf('klassenkarte') != -1) {
            var textarea = forms[n].getElementsByTagName('textarea')[0];
            var select = forms[n].getElementsByTagName('select')[0];
            select.innerHTML = '';
            for (var z = 0; z < this._classes.length; ++z)
                ScoTools.createHtml('option', select, null, this._classes[z].className);
            var fktChange = function() {
                textarea.value = self._classes[select.selectedIndex].builder.getAttributeNames().join('\n');
            }
            fktChange();
            select.addEventListener('change', fktChange);
        }
        if (classes.indexOf('objektkarte') != -1) {
            this._okt = forms[n].getElementsByTagName('input')[0];
            this._okt.value = data.title;
            this._oka = forms[n].getElementsByTagName('textarea')[0];
            this._oka.value = data.atts;
            forms[n].addEventListener('input', function() { self.procChange(); });
            forms[n].addEventListener('change', function() { self.doSave(); });
        }
    }
}

/**
* Einen menschenlesbaren String mit noch nötiger Objektkartenanzahl erstellen
* @return String    Erstellter String
*/
ReadClasscardTask.prototype.howMuchLeftString = function() {
    if (this.countDone >= this.countNeeded)
        return 'Natürlich kannst du weiterüben, obwohl du diese Aufgabe bereits erfolgreich abgeschlossen hast';
    return 'Fülle die Objektkarte so aus, dass sie das Zielobjekt beschreibt';
}

/**
* Festsetzen, ob die Werte in der Objektkarte gültig sind und ggf. Fehlermeldung anzeigen
* @return undefined
*/
ReadClasscardTask.prototype._reportError = function(
    strMsg,         // Anzuzeigende Meldung
    boolTitle       // Optional: Ist der Titel das Problem?
) {
    if (strMsg)
        this.showMsg(strMsg, ScoTools.MSGTYPE_WARN);
    if (this._okt.setCustomValidity)
        this._okt.setCustomValidity((!strMsg || !boolTitle) ? '' : strMsg);
    if (this._oka.setCustomValidity)
        this._oka.setCustomValidity((!strMsg || boolTitle) ? '' : strMsg);
}

/**
* Darauf reagieren, wenn Objektkarte verändert wird
* @return undefined
*/
ReadClasscardTask.prototype.procChange = function() {
    var title = this._okt.value;
    var atts = this._oka.value;

    this.showObject();
    // Objektkarte leer?
    if (!(title + atts).length) {
        this.showMsg(this.howMuchLeftString(), ScoTools.MSGTYPE_INFO);
        return;
    }
    // Titelzeile OK?
    var title = ReadClasscardTask.parseTitle(title);
    if (title.error) {
        this._reportError(title.error, true);
        return;
    }
    var classIndex = this._classes.map(function(item) { return item.className; }).indexOf(title.klasse);
    if (classIndex == -1) {
        this._reportError('Angegebene Klasse „' + title.klasse + '“ ist unbekannt', true);
        return;
    }
    // Attribute prinzipiell OK?
    atts = ReadClasscardTask.parseAttributes(atts, this._classes[classIndex].builder.getAttributeNames());
    if ((typeof atts == 'string') || (atts instanceof String)) {
        this._reportError(atts);
        return;
    }
    // Datenobjekt erstellen
    var parsingResult = this._classes[classIndex].builder.createFromAtts(atts);
    if (parsingResult.message) {
        this._reportError(parsingResult.message);
        return;
    }
    // Alles als Fehlerfrei markieren
    this._reportError();
    this.showObject(title.klasse, parsingResult.object);

    // Passt die Objektkarte komplett?
    if (this.isTarget(parsingResult.object, title.klasse))
        this.reportCorrect();
    else
        this.showMsg('Die Objektkarte passt nicht zum Zielobjekt…');
}

/**
* Prüft, ob die eingegebenen Attributwerte atts mit einer Zielkonfiguration übereinstimmt
* @return Boolean   Übereinstimmung?
*/
ReadClasscardTask.prototype.isTarget = function(
    atts,       // Tatsächliche Attributwerte aus Objektkarte
    className   // Klassenname aus Objektkarte
) {
    return this.targetConfig.reduce(function(isTarget, target) {
        return isTarget || ReadClasscardTask.isEqual(atts, className, target);
    }, false);
}

/**
* Prüft, ob die eingegebenen Attributwerte atts mit der Zielkonfiguration target übereinstimmt
* @return Boolean   Übereinstimmung?
*/
ReadClasscardTask.isEqual = function(
    atts,       // Tatsächliche Attributwerte aus Objektkarte
    className,  // Klassenname aus Objektkarte
    target      // Geforderte Zielkonfiguration
) {
    for (var key in target) {
        if (key == 'className') {
            if (className != target['className'])
                return false;
        } else if (target[key] instanceof Array)
            return target[key].map((value, index) => (value == atts[key][index])).reduce((a, b) => (a && b), true);
        else if (atts[key] != target[key])
            return false;
    }
    return true;
}

/**
* Eine Meldung über das Ergebnis der Antwortüberprüfung anzeigen
* @return undefined
*/
ReadClasscardTask.prototype.reportCorrect = function() {
    ++this.countDone;
    // Erfolgsmeldung
    this.showMsg((this.countDone < this.countNeeded) ? ('Richtig: ' + this.howMuchLeftString()) : 'Du hast die Objektkarte fehlerfrei ausgefüllt', true);
    this.showFinalMessage();
    // Neue Aufgabe
    /*!*//*
    this._okt.value = '';
    this._oka.value = '';
    this._reportError();*/
    // Speichern
    this.doSave();
    // Erfolgssound
    if (this.ready)
        this.playAudio(true);
    // Titelzeile fokusieren
    //this._okt.focus();
}
